from contextlib import contextmanager
import logging
import re
from typing import List, Optional, Dict, Union
from enum import Enum

import paramiko

default_log = logging.getLogger(__name__)

TOP_PROCESSES_COUNT = 20
COMMAND_REPORT_INTERVAL = 5 # some commands need you to specify an interval e.g. vmstat 1 2

def to_bytes(value: str) -> int:
    value = value.upper()
    if re.search(r'[bkmgtBKMGT]$', value):
        unit = value[-1]
        float_value = float(re.sub("B|K|M|G|T", "", value))
        if unit == "B":
            return float_value
        if unit == "M":
            return float_value * 10 ** 6
        if unit == "G":
            return float_value * 10 ** 9
        if unit == "T":
            return float_value * 10 ** 12
        return float_value * 10 ** 3
    else:
        return float(value)

def chunks(lst, n):
    for i in range(0, len(lst), n):
        yield lst[i:i + n]

class OS(Enum):
    SOLARIS = "Solaris"
    SUSE = "Suse"
    LINUX = "Generic Linux"
    RHEL = "RHEL"
    AIX = "AIX"
    CENTOS = "Centos"
    FREE_BSD = "FreeBSD"
    ORACLE_LINUX = "Oracle Linux"

class NodeStatus(Enum):
    UP = "U"
    DOWN = "D"

class NodeState(Enum):
    NORMAL = "N"
    LEAVING = "L"
    JOINING = "J"
    MOVING = "M"

enum_translation = {
    "U": "UP",
    "D": "DOWN",
    "N": "NORMAL",
    "L": "LEAVING",
    "J": "JOINING",
    "M": "MOVING"
}

# these aren't perfectly accurate names for the lists but mainly just say if we can use sysstat tools or if
# sources like /proc/stat etc... will be available (standard has these files)
STANDARD_LINUX = [OS.LINUX, OS.CENTOS, OS.RHEL, OS.SUSE, OS.ORACLE_LINUX]
PROPRIETARY_LINUX = [OS.SOLARIS, OS.AIX, OS.FREE_BSD]

class Node:
    def __init__(self, status: NodeStatus, state: NodeState, address: str, load: float, tokens: int, owns: Union[None,float], host_id: str, rack: str) -> None:
        self.status = enum_translation.get(status.value)
        self.state = enum_translation.get(state.value)
        self.address = address
        self.load = load
        self.tokens = tokens
        self.owns = owns
        self.host_id = host_id
        self.rack = rack

class InfoResult:
    def __init__(self, lines: List[str]) -> None:
        """
        ID                     : 341c47b4-7f39-4733-a424-87c24f4c4b11
        Gossip active          : true
        Native Transport active: true
        Load                   : 172.12 KiB
        Generation No          : 1641405003
        Uptime (seconds)       : 81966
        Heap Memory (MB)       : 235.54 / 1004.00
        Off Heap Memory (MB)   : 0.00
        Data Center            : datacenter1
        Rack                   : rack1
        Exceptions             : 0
        Key Cache              : entries 20, size 1.95 KiB, capacity 50 MiB, 87 hits, 110 requests, 0.791 recent hit rate, 14400 save period in seconds
        Row Cache              : entries 0, size 0 bytes, capacity 0 bytes, 0 hits, 0 requests, NaN recent hit rate, 0 save period in seconds
        Counter Cache          : entries 0, size 0 bytes, capacity 25 MiB, 0 hits, 0 requests, NaN recent hit rate, 7200 save period in seconds
        Percent Repaired       : 100.0%
        Token                  : -8850987814906043538
        Token                  : -7607467809225451564
        """
        token_count = 0
        for line in lines:
            if line.startswith("ID"):
                _, host_id = line.split(": ")
            elif line.startswith("Gossip active"):
                _, gossip_active = line.split(": ")
            elif line.startswith("Native Transport"):
                _, native_transport_active = line.split(": ")
            elif line.startswith("Load"):
                _, load_and_unit = line.split(": ")
                load, load_unit = load_and_unit.split()
            elif line.startswith("Uptime"):
                _, uptime = line.split(": ")
            elif line.startswith("Heap Memory"):
                _, fraction = line.split(": ")
                heap_used, heap_total = fraction.split(" / ")
            elif line.startswith("Off Heap Memory"):
                continue
            elif line.startswith("Data Center"):
                _, datacenter = line.split(": ")
            elif line.startswith("Rack"):
                _, rack = line.split(": ")
            elif line.startswith("Exceptions"):
                _, exceptions = line.split(": ")
            elif line.startswith("Key Cache"):
                _, key_cache_line = line.split(": ")
                _, key_cache_entries, _, key_cache_size, key_cache_size_unit, _, key_cache_capacity, key_cache_capacity_unit, key_cache_hits, _, key_cache_requests, _, key_cache_recent_hit_rate, _, _, _, key_cache_save_period_in_seconds, *_ = key_cache_line.replace(",", "").split()
            elif line.startswith("Row Cache"):
                _, row_cache_line = line.split(": ")
                _, row_cache_entries, _, row_cache_size, row_cache_size_unit, _, row_cache_capacity, row_cache_capacity_unit, row_cache_hits, _, row_cache_requests, _, row_cache_recent_hit_rate, _, _, _, row_cache_save_period_in_seconds, *_ = row_cache_line.replace(",", "").split()
            elif line.startswith("Counter Cache"):
                _, counter_cache_line = line.split(": ")
                _, counter_cache_entries, _, counter_cache_size, counter_cache_size_unit, _, counter_cache_capacity, counter_cache_capacity_unit, counter_cache_hits, _, counter_cache_requests, _, counter_cache_recent_hit_rate, _, _, _, counter_cache_save_period_in_seconds, *_ = counter_cache_line.replace(",", "").split()
            elif line.startswith("Percent Repaired"):
                _, percent_repaired = line.split(": ")
            elif line.startswith("Token"):
                token_count += 1

        
        self.host_id = host_id
        self.gossip_active = True if gossip_active == "true" else False
        self.native_transport_active = True if native_transport_active == "true" else False
        self.load = to_bytes(f"{load}{load_unit[0]}")
        self.uptime = int(uptime)
        self.heap_used = to_bytes(f"{heap_used}M")
        self.heap_total = to_bytes(f"{heap_total}M")
        self.heap_used_percentage = (self.heap_used / self.heap_total) * 100
        self.datacenter = datacenter
        self.rack = rack
        self.exceptions = int(exceptions)

        # key cache
        self.key_cache_entries = int(key_cache_entries)
        self.key_cache_size = to_bytes(f"{key_cache_size}{key_cache_size_unit[0]}")
        self.key_cache_capacity = to_bytes(f"{key_cache_capacity}{key_cache_capacity_unit[0]}")
        self.key_cache_hits = int(key_cache_hits)
        self.key_cache_requests = int(key_cache_requests)
        self.key_cache_hit_rate = (float(key_cache_recent_hit_rate)) * 100 if key_cache_recent_hit_rate != "NaN" else 0
        # row cache
        self.row_cache_entries = int(row_cache_entries)
        self.row_cache_size = to_bytes(f"{row_cache_size}{row_cache_size_unit[0]}")
        self.row_cache_capacity = to_bytes(f"{row_cache_capacity}{row_cache_capacity_unit[0]}")
        self.row_cache_hits = int(row_cache_hits)
        self.row_cache_requests = int(row_cache_requests)
        self.row_cache_hit_rate = (float(row_cache_recent_hit_rate)) * 100 if row_cache_recent_hit_rate != "NaN" else 0
        # counter cache
        self.counter_cache_entries = int(counter_cache_entries)
        self.counter_cache_size = to_bytes(f"{counter_cache_size}{counter_cache_size_unit[0]}")
        self.counter_cache_capacity = to_bytes(f"{counter_cache_capacity}{counter_cache_capacity_unit[0]}")
        self.counter_cache_hits = int(counter_cache_hits)
        self.counter_cache_requests = int(counter_cache_requests)
        self.counter_cache_hit_rate = (float(counter_cache_recent_hit_rate)) * 100 if counter_cache_recent_hit_rate != "NaN" else 0

        self.tokens = token_count

class StatusResult:
    def __init__(self, lines: List[str]) -> None:
        """
        Datacenter: datacenter1
        =======================
        Status=Up/Down
        |/ State=Normal/Leaving/Joining/Moving
        --  Address    Load       Tokens  Owns    Host ID                               Rack
        UN  127.0.0.1  47.66 KB   1       33.3%   aaa1b7c1-6049-4a08-ad3e-3697a0e30e10  rack1
        UN  127.0.0.2  47.67 KB   1       33.3%   1848c369-4306-4874-afdf-5c1e95b8732e  rack1
        UN  127.0.0.3  47.67 KB   1       33.3%   49578bf1-728f-438d-b1c1-d8dd644b6f7f  rack1
        Datacenter: DC2
        ===============
        Status=Up/Down
        |/ State=Normal/Leaving/Joining/Moving
        --  Address         Load       Tokens       Owns    Host ID                               Rack
        UN  10.128.181.181  64.77 GiB  256          ?       fa134466-eb59-47d9-adee-c69eb23027c3  RAC1
        """
        self.datacenters = {}
        nodes: List[Node] = []
        begin_node_lines = False
        for line in lines:
            if line.startswith("Datacenter:"):
                _, datacenter = line.split()
                self.datacenters.update({datacenter: []})
                begin_node_lines = False
                nodes = []
            elif line.startswith("--"):
                begin_node_lines = True
            elif begin_node_lines:
                try:
                    status_and_state, address, load, unit, tokens, owns, host_id, rack = line.split()
                    status = NodeStatus(status_and_state[0])
                    state = NodeState(status_and_state[1])
                    try:
                        owns = float(owns.replace("%", ""))
                    except Exception as e:
                        owns = None
                    self.datacenters[datacenter].append(Node(status, state, address, to_bytes(f"{load}{unit[0]}"), int(tokens), owns, host_id, rack))
                except Exception as e:
                    default_log.debug(f"Status parser - skipping line: {line}")

class RemoteCassandraHost:
    def __init__(
        self, os: OS, hostname: str, username: str, port=22, password: Optional[str] = None, private_key=None, private_key_password=None, cassandra_user="", cassandra_password="", nodetool_path="nodetool", cassandra_port="", log=default_log,
    ):
        self.os = os
        self.hostname = hostname
        self.username = username
        self.password = password
        self.port = port
        self.cassandra_user = cassandra_user
        self.cassandra_password = cassandra_password
        self.cassandra_port = cassandra_port
        if self.cassandra_port != "":
            self.port_arg = f" --port {self.cassandra_port}"
        else:
            self.port_arg = ""
        self.nodetool_command_base = f"{nodetool_path}{self.port_arg}" if self.cassandra_user == "" else f"{nodetool_path}{self.port_arg} --username {self.cassandra_user} --password {self.cassandra_password}"

        self._log = log
        self._private_key = private_key
        self._private_key_password = private_key_password
        self._client = paramiko.SSHClient()
        self._client.load_system_host_keys()
        self._client.set_missing_host_key_policy(paramiko.AutoAddPolicy)

    def run_command(self, cmd: str) -> List[str]:
        try:
            stdin, stdout, stderr = self._client.exec_command(cmd, timeout=30)
            err = stderr.read().decode("utf-8")
            out = stdout.read().decode("utf-8")
            if err:
                self._log.warning(f"command stderr: {err}")
            lines = [l.strip() for l in out.splitlines() if l.strip()]
            self._log.debug(f"{self.hostname} - command output lines: {lines}")
            return lines
        except Exception as e:
            self._log.exception(f"{self.hostname} - error running command: {e}")
            return []

    @contextmanager
    def connect(self):
        if self._private_key is not None:
            self._log.info(f"{self.hostname} - attempting to connect using key: {self._private_key}, port: {self.port}, username: {self.username}")
            self._client.connect(
                self.hostname, port=self.port, username=self.username, key_filename=self._private_key, passphrase=self._private_key_password,
            )
        else:
            self._log.info(f"{self.hostname} - attempting to connect using password, port: {self.port}, username: {self.username}")
            self._client.connect(self.hostname, port=self.port, username=self.username, password=self.password, timeout=20)
        yield self._client
        if self._client:
            self._client.close()
            self._log.info(f"{self.hostname} - disconnecting")

    def run_nodetool_info(self):
        nodetool_info = InfoResult(self.run_command(f"{self.nodetool_command_base} info -T"))
        return nodetool_info

    def run_nodetool_status(self):
        nodetool_status = StatusResult(self.run_command(f"{self.nodetool_command_base} status"))
        return nodetool_status

    def get_cassandra_version(self) -> str:
        _, version = self.run_command(f"{self.nodetool_command_base} version")[0].split()
        return version




