import logging
import functools
import time
from concurrent.futures import ThreadPoolExecutor
from typing import List

from ruxit.api.base_plugin import RemoteBasePlugin
from cassandra_api import RemoteCassandraHost, OS

log = logging.getLogger(__name__)

def time_it(function):
    """
    Annotation used to log method execution times
    """

    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        start = time.time()
        ret = function(*args, **kwargs)
        delta = time.time() - start

        args[0].logger.info(f"{args[0].config.get('hostname')} - execution of {function.__name__} took {delta:.3f} seconds")

        return ret

    return wrapper

def exception_logger(function):
    """
    Annotation used to log method exceptions
    """

    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except Exception as e:
            args[0].logger.exception(f"{args[0].config.get('hostname')} - error executing {function.__name__}: '{e}'")
            raise

    return wrapper

class CassandraRemoteExtension(RemoteBasePlugin):
    def initialize(self, **kwargs):
        self.execution_count = 0
        self.had_successfull_run = False # used to decide if a connection error should throw and error on UI or just be reported as an outage
        self.result_store = {}
        self.cluster_status = None
        self.report_only_local_node = False

    @time_it
    def query(self, **kwargs):

        log.setLevel(self.config.get("log_level", "INFO"))

        alias = self.config.get("alias") if self.config.get("alias") else self.config.get("hostname")

        group = self.topology_builder.create_group(self.config.get("group"), self.config.get("group"))
        self.device = group.create_device(alias)

        password = self.config.get("password") if self.config.get("password") else None
        key = self.config.get("ssh_key_file") if self.config.get("ssh_key_file") else None
        passphrase = self.config.get("ssh_key_passphrase") if self.config.get("ssh_key_passphrase") else None

        self.remote_cassandra = RemoteCassandraHost(
            OS(self.config.get("os", None)),
            self.config.get("hostname"),
            self.config.get("username"),
            self.config.get("port", 22),
            password=password,
            private_key=key,
            private_key_password=passphrase,
            cassandra_user=self.config.get("cassandra_username", ""),
            cassandra_password=self.config.get("cassandra_password", ""),
            cassandra_port=self.config.get("cassandra_port", ""),
            nodetool_path=self.config.get("nodetool_path", "nodetool"),
            log=log,
        )

        self.should_report_properties = True if ((self.execution_count % 30 == 0) or self.execution_count == 0) else False

        if self.should_report_properties:
            self.report_additional_props()

        try:
            host_is_available = True
            with self.remote_cassandra.connect():
                try:
                    self.cluster_status = self.remote_cassandra.run_nodetool_status()
                except Exception as e:
                    raise ConnectionError(e)
                if self.should_report_properties:
                    self.device.report_property("datacenters", ",".join([dc for dc in self.cluster_status.datacenters]))

                with ThreadPoolExecutor(max_workers=5) as e:
                    e.submit(self.report_cluster_status)
                    e.submit(self.report_node_info)

                    if self.should_report_properties:
                        e.submit(self.report_version)
                        
        except Exception as e:
            host_is_available = False
            if self.had_successfull_run:
                self.device.report_custom_info_event(f"{e}", "Error connecting to host." )
                log.warning(f"Error connecting to {self.config.get('hostname')}: {e}")
            else:
                raise ConnectionError(f"Initial error on connection to {self.config.get('hostname')}: {e}")

        self.device.absolute("availability", 1 if host_is_available else 0)    
        self.had_successfull_run = True
        self.execution_count += 1

    def report_additional_props(self):
        properties = self.config.get("additional_props", "").split("\n") if self.config.get ("additional_props") else []
        for property in properties:
            try:
                property = property.strip()
                key, value = property.split("=")
                self.device.report_property(key, value)
            except Exception as e:
                log.warn(f"{self.config.get('hostname')} - Unable to report property {property}: e")

    @time_it
    @exception_logger
    def report_node_info(self):
        node = self.remote_cassandra.run_nodetool_info()
        self.device.absolute("gossip_active", node.gossip_active, dimensions={"host_id": node.host_id, "rack": node.rack, "datacenter": node.datacenter})
        self.device.absolute("native_transport_active", node.native_transport_active, dimensions={"host_id": node.host_id, "rack": node.rack, "datacenter": node.datacenter})
        self.device.absolute("node_load", node.load, dimensions={"host_id": node.host_id, "rack": node.rack, "datacenter": node.datacenter})
        self.device.absolute("uptime", node.uptime, dimensions={"host_id": node.host_id, "rack": node.rack, "datacenter": node.datacenter})
        self.device.absolute("heap_used", node.heap_used, dimensions={"host_id": node.host_id, "rack": node.rack, "datacenter": node.datacenter})
        self.device.absolute("heap_total", node.heap_total, dimensions={"host_id": node.host_id, "rack": node.rack, "datacenter": node.datacenter})
        self.device.absolute("heap_used_percentage", node.heap_used_percentage, dimensions={"host_id": node.host_id, "rack": node.rack, "datacenter": node.datacenter})
        self.device.relative("exceptions", node.exceptions, dimensions={"host_id": node.host_id, "rack": node.rack, "datacenter": node.datacenter})

        self.device.absolute("key_cache_entries", node.key_cache_entries, dimensions={"host_id": node.host_id, "rack": node.rack, "datacenter": node.datacenter})
        self.device.absolute("key_cache_size", node.key_cache_size, dimensions={"host_id": node.host_id, "rack": node.rack, "datacenter": node.datacenter})
        self.device.absolute("key_cache_capacity", node.key_cache_capacity, dimensions={"host_id": node.host_id, "rack": node.rack, "datacenter": node.datacenter})
        self.device.absolute("key_cache_hits", node.key_cache_hits, dimensions={"host_id": node.host_id, "rack": node.rack, "datacenter": node.datacenter})
        self.device.absolute("key_cache_requests", node.key_cache_requests, dimensions={"host_id": node.host_id, "rack": node.rack, "datacenter": node.datacenter})
        self.device.absolute("key_cache_hit_rate", node.key_cache_hit_rate, dimensions={"host_id": node.host_id, "rack": node.rack, "datacenter": node.datacenter})

        self.device.absolute("row_cache_entries", node.row_cache_entries, dimensions={"host_id": node.host_id, "rack": node.rack, "datacenter": node.datacenter})
        self.device.absolute("row_cache_size", node.row_cache_size, dimensions={"host_id": node.host_id, "rack": node.rack, "datacenter": node.datacenter})
        self.device.absolute("row_cache_capacity", node.row_cache_capacity, dimensions={"host_id": node.host_id, "rack": node.rack, "datacenter": node.datacenter})
        self.device.absolute("row_cache_hits", node.row_cache_hits, dimensions={"host_id": node.host_id, "rack": node.rack, "datacenter": node.datacenter})
        self.device.absolute("row_cache_requests", node.row_cache_requests, dimensions={"host_id": node.host_id, "rack": node.rack, "datacenter": node.datacenter})
        self.device.absolute("row_cache_hit_rate", node.row_cache_hit_rate, dimensions={"host_id": node.host_id, "rack": node.rack, "datacenter": node.datacenter})

        self.device.absolute("counter_cache_entries", node.counter_cache_entries, dimensions={"host_id": node.host_id, "rack": node.rack, "datacenter": node.datacenter})
        self.device.absolute("counter_cache_size", node.counter_cache_size, dimensions={"host_id": node.host_id, "rack": node.rack, "datacenter": node.datacenter})
        self.device.absolute("counter_cache_capacity", node.counter_cache_capacity, dimensions={"host_id": node.host_id, "rack": node.rack, "datacenter": node.datacenter})
        self.device.absolute("counter_cache_hits", node.counter_cache_hits, dimensions={"host_id": node.host_id, "rack": node.rack, "datacenter": node.datacenter})
        self.device.absolute("counter_cache_requests", node.counter_cache_requests, dimensions={"host_id": node.host_id, "rack": node.rack, "datacenter": node.datacenter})
        self.device.absolute("counter_cache_hit_rate", node.counter_cache_hit_rate, dimensions={"host_id": node.host_id, "rack": node.rack, "datacenter": node.datacenter})

        self.device.absolute("node_tokens", node.tokens, dimensions={"host_id": node.host_id, "rack": node.rack, "datacenter": node.datacenter})

        for datacenter in self.cluster_status.datacenters:
            for cluster_node in self.cluster_status.datacenters[datacenter]:
                if cluster_node.host_id == node.host_id:
                    self.device.absolute("node_ownership", cluster_node.owns, dimensions={"datacenter": datacenter, "rack": cluster_node.rack, "host_id": cluster_node.host_id})
                    self.device.state_metric("node_state", cluster_node.state, dimensions={"datacenter": datacenter, "rack": cluster_node.rack, "host_id": cluster_node.host_id})
                    self.device.state_metric("node_status", cluster_node.status, dimensions={"datacenter": datacenter, "rack": cluster_node.rack, "host_id": cluster_node.host_id})

    @time_it
    @exception_logger
    def report_cluster_status(self):
        if self.cluster_status and self.config.get("report_cluster_info", True):
            for datacenter in self.cluster_status.datacenters:
                for node in self.cluster_status.datacenters[datacenter]:
                    if node.owns is not None:
                        self.device.absolute("cluster_node_ownership", node.owns, dimensions={"datacenter": datacenter, "address": node.address, "rack": node.rack, "host_id": node.host_id})
                    self.device.absolute("cluster_node_load", node.load, dimensions={"datacenter": datacenter, "address": node.address, "rack": node.rack, "host_id": node.host_id})
                    self.device.absolute("cluster_node_tokens", node.tokens, dimensions={"datacenter": datacenter, "address": node.address, "rack": node.rack, "host_id": node.host_id})
                    self.device.state_metric("cluster_node_state", node.state, dimensions={"datacenter": datacenter, "address": node.address, "rack": node.rack, "host_id": node.host_id})
                    self.device.state_metric("cluster_node_status", node.status, dimensions={"datacenter": datacenter, "address": node.address, "rack": node.rack, "host_id": node.host_id})
        else:
            log.info(f"Not reporting cluster wide cluster level metrics.")

    @time_it
    @exception_logger
    def report_version(self):
        version = self.remote_cassandra.get_cassandra_version()
        self.device.report_property("cassandra_version", version)
